/*++
 
// THIS CODE AND INFORMATION IS PROVIDED "AS IS" WITHOUT WARRANTY OF
// ANY KIND, EITHER EXPRESSED OR IMPLIED, INCLUDING BUT NOT LIMITED TO
// THE IMPLIED WARRANTIES OF MERCHANTABILITY AND/OR FITNESS FOR A
// PARTICULAR PURPOSE.

Module Name:

    SensorDDI.cpp

Abstract:

      This module implements the ISensorDriver interface which is used
      by the Sensor Class Extension.
--*/

#include "internal.h"
#include "Defines.h"
#include "SensorDriver.h"
#include "WpdObjectProperties.h"
#include "ReadWriteRequest.h"
#include "SensorDDI.h"

#include "SensorDDI.tmh"

/////////////////////////////////////////////////////////////////////////
//
// CSensorDDI::CSensorDDI()
//
// Constructor.
//
/////////////////////////////////////////////////////////////////////////
CSensorDDI::CSensorDDI() :
    m_dwClients(0)
{
    // Initialize HID Read/Write requests to NULL
    for( DWORD i = 0; i < PENDING_READ_COUNT; i++ )
    {
        m_pHidReadRequest[i] = NULL;
    }

    m_pHidWriteRequest = NULL;
}

/////////////////////////////////////////////////////////////////////////
//
// CSensorDDI::~CSensorDDI()
//
// Destructor
//
/////////////////////////////////////////////////////////////////////////
CSensorDDI::~CSensorDDI()
{
    for( DWORD i = 0; i < PENDING_READ_COUNT; i++ )
    {
        SAFE_RELEASE(m_pHidReadRequest[i]);
    }

    SAFE_RELEASE(m_pHidWriteRequest);
}

/////////////////////////////////////////////////////////////////////////
//
// CSensorDDI::InitSensorDevice
//
// 
//
/////////////////////////////////////////////////////////////////////////
HRESULT CSensorDDI::InitSensorDevice(__in IWDFDevice* pWdfDevice)
{
    Trace(TRACE_LEVEL_INFORMATION, "%!FUNC! Entry");

    HRESULT hr = (NULL != pWdfDevice) ? S_OK : E_UNEXPECTED;

    // Create the HID Read objects
    // Initialize the request with the IWDFDevice object
    // Send the pending read request that will get completed
    if(SUCCEEDED(hr))
    {
        for( DWORD i = 0; SUCCEEDED(hr) && i < PENDING_READ_COUNT; i++ )
        {
            hr = CComObject<CReadWriteRequest>::CreateInstance(&m_pHidReadRequest[i]);
            if ((SUCCEEDED(hr)) && (NULL != m_pHidReadRequest[i]))
            {
                m_pHidReadRequest[i]->AddRef();

                hr = m_pHidReadRequest[i]->InitializeRequest(pWdfDevice, this);

                if(SUCCEEDED(hr))
                {
                    hr = m_pHidReadRequest[i]->CreateAndSendReadRequest();
                }
                else
                {
                    break;
                }
            }
            else
            {
                break;
            }
        }
       
        // Create the HID Write objects
        // Initialize the request with the IWDFDevice object
        hr = CComObject<CReadWriteRequest>::CreateInstance(&m_pHidWriteRequest);
        if ((SUCCEEDED(hr)) && (NULL != m_pHidWriteRequest))
        {
            m_pHidWriteRequest->AddRef();

            hr = m_pHidWriteRequest->InitializeRequest(pWdfDevice, NULL);
        }

        // Set default startup values


    }

    return hr;
}

/////////////////////////////////////////////////////////////////////////
//
// CSensorDDI::DeInitSensorDevice
//
// 
//
/////////////////////////////////////////////////////////////////////////
HRESULT CSensorDDI::DeInitSensorDevice()
{
    Trace(TRACE_LEVEL_INFORMATION, "%!FUNC! Entry");

    HRESULT hr = S_OK;

    for( DWORD i = 0; i < PENDING_READ_COUNT; i++ )
    {
        m_pHidReadRequest[i]->CancelAndStopPendingRequest();
        m_pHidReadRequest[i]->UninitializeRequest();
        
        SAFE_RELEASE(m_pHidReadRequest[i]);
    }

    m_pHidWriteRequest->CancelAndStopPendingRequest();
    m_pHidWriteRequest->UninitializeRequest();
    
    SAFE_RELEASE(m_pHidWriteRequest);

    return hr;
}

/////////////////////////////////////////////////////////////////////////
//
// CSensorDDI::OnGetSupportedSensorObjects
//
// Routine Description:
//
//  This method is called by Sensor Class Extension during the initialize procedure to get 
//  the list of sensor objects and their supported properties.
//  
//  NOTE:   This driver supports 4 sensors:
//	        one accelerometer
//          one light sensor
//          two switch arrays
//
//
// Arguments:
//
//  ppPortableDeviceValuesCollection - A double IPortableDeviceValuesCollection pointer
//                                     that receives the set of Sensor property values
//
// Return Value:
//
//  Status
// 
//
/////////////////////////////////////////////////////////////////////////
HRESULT CSensorDDI:: OnGetSupportedSensorObjects(
        __out IPortableDeviceValuesCollection** ppPortableDeviceValuesCollection
        )
{
    HRESULT hr = S_OK;
	
    // CoCreate a collection to store the sensor object identifiers.
    hr = CoCreateInstance(CLSID_PortableDeviceValuesCollection,
                              NULL,
                              CLSCTX_INPROC_SERVER,
                              IID_IPortableDeviceValuesCollection,
                              (VOID**) ppPortableDeviceValuesCollection);

    CComBSTR objectId;
    CComPtr<IPortableDeviceKeyCollection> pKeys;
    CComPtr<IPortableDeviceValues> pValues;

	POSITION pos = g_SupportedSensors.GetStartPosition();

    while( NULL != pos)
    {
        objectId = g_SupportedSensors.GetNextValue(pos);


	    if (hr == S_OK)
		{
			hr = OnGetSupportedProperties(objectId, &pKeys);
		}

		if (hr == S_OK)
		{
			hr = m_properties.OnGetPropertyValues( objectId, pKeys, &pValues );
		}

		if (hr == S_OK)
		{
			hr = (*ppPortableDeviceValuesCollection)->Add(pValues);
		}

		pKeys.Release();
		pValues.Release();
	}

    return hr;
}

/////////////////////////////////////////////////////////////////////////
//
// CSensorDDI::OnGetSupportedProperties
//
// Routine Description:
//
//  This method is called by Sensor Class Extension to get the list of supported properties
//  for a particular Sensor.
//  
// Arguments:
//
//  ObjectID - String that represents the object whose supported property keys are being requested
//  ppKeys - An IPortableDeviceKeyCollection to be populated with supported PROPERTYKEYs
//
// Return Value:
//  Status
// 
//
/////////////////////////////////////////////////////////////////////////
HRESULT CSensorDDI:: OnGetSupportedProperties(
        __in  LPWSTR ObjectID,
        __out IPortableDeviceKeyCollection** ppKeys
        )
{
    HRESULT hr = S_OK;

    hr = m_properties.OnGetSupportedProperties( ObjectID, ppKeys );

    return hr;
}

/////////////////////////////////////////////////////////////////////////
//
// CSensorDDI::OnGetSupportedDataFields
//
// Routine Description:
//
//  This method is called by Sensor Class Extension to get the list of supported data fields
//  for a particular Sensor.
//  
// Arguments:
//
//  ObjectID - String that represents the object whose supported property keys are being requested
//  ppKeys - An IPortableDeviceKeyCollection to be populated with supported PROPERTYKEYs
//
// Return Value:
//  Status
// 
//
/////////////////////////////////////////////////////////////////////////
HRESULT CSensorDDI:: OnGetSupportedDataFields(
        __in  LPWSTR ObjectID,
        __out IPortableDeviceKeyCollection** ppKeys
        )
{
    HRESULT hr = S_OK;

    hr = m_properties.OnGetSupportedDataFields( ObjectID, ppKeys );

    return hr;
}

/////////////////////////////////////////////////////////////////////////
//
// CSensorDDI::OnGetSupportedEvents
//
// Routine Description:
//
//  This method is called by Sensor Class Extension to get the list of supported events
//  for a particular Sensor.
//  
// Arguments:
//
// ObjectID - String that represents the object whose supported property keys are being requested
//  ppSupportedEvents - A set of GUIDs that represent the supported events
//  pulEventCount - Count of the number of events supported
//
// Return Value:
//  Status
// 
//
/////////////////////////////////////////////////////////////////////////
HRESULT CSensorDDI:: OnGetSupportedEvents(
        __in  LPWSTR ObjectID,
        __out GUID **ppSupportedEvents,
        __out ULONG *pulEventCount
        )
{
    UNREFERENCED_PARAMETER(ObjectID);

    HRESULT hr = S_OK;

    //Return two GUIDs
    GUID* pBuf = (GUID*)CoTaskMemAlloc(sizeof(GUID) * 2);

    if( pBuf != NULL )
    {
        *pBuf       = SENSOR_EVENT_DATA_UPDATED;
        *(pBuf + 1) = SENSOR_EVENT_STATE_CHANGED;

        *ppSupportedEvents = pBuf;
        *pulEventCount = 2;
    }
    else
    {
        hr = E_OUTOFMEMORY;

        *ppSupportedEvents = NULL;
        *pulEventCount = 0;
    }

    return hr;
}

/////////////////////////////////////////////////////////////////////////
//
// CSensorDDI::OnGetProperties
//
// Routine Description:
//
//  This method is called by Sensor Class Extension to get Sensor property values 
//  for a particular Sensor.
//  
// Arguments:
//
//  ObjectID - String that represents the object whose supported property keys are being requested
// 
//
// Return Value:
//  Status
// 
//
/////////////////////////////////////////////////////////////////////////
HRESULT CSensorDDI:: OnGetProperties(
        __in  IWDFFile* appId,
        __in  LPWSTR ObjectID,
        __in  IPortableDeviceKeyCollection* pKeys,
        __out IPortableDeviceValues** ppValues
        )
{
    UNREFERENCED_PARAMETER(appId);

    HRESULT hr = S_OK;

    if ((ObjectID == NULL) ||
        (pKeys       == NULL) ||
        (ppValues     == NULL))
    {
        hr = E_INVALIDARG;
        return hr;
    }

    hr = m_properties.OnGetPropertyValues( ObjectID, pKeys, ppValues );

    return hr;
}

/////////////////////////////////////////////////////////////////////////
//
// CSensorDDI::OnGetDataFields
//
// Routine Description:
//
//  This method is called by Sensor Class Extension to get Sensor data fields
//  for a particular Sensor.
//  
// Arguments:
//
//  ObjectID - String that represents the object whose supported property keys are being requested
// 
//
// Return Value:
//  Status
// 
//
/////////////////////////////////////////////////////////////////////////
HRESULT CSensorDDI:: OnGetDataFields(
        __in  IWDFFile* appId,
        __in  LPWSTR ObjectID,
        __in  IPortableDeviceKeyCollection* pKeys,
        __out IPortableDeviceValues** ppValues
        )
{
    UNREFERENCED_PARAMETER(appId);

    HRESULT hr = S_OK;

    if ((ObjectID == NULL) ||
        (pKeys       == NULL) ||
        (ppValues     == NULL))
    {
        hr = E_INVALIDARG;
        return hr;
    }

    hr = m_properties.OnGetDataFields( ObjectID, pKeys, ppValues );

    return hr;
}

/////////////////////////////////////////////////////////////////////////
//
// CSensorDDI::OnSetProperties
//
// Routine Description:
//
//  This method is called by Sensor Class Extension to set Sensor properties
//  for a particular Sensor.
//  
// Arguments:
//
//  ObjectID - String that represents the object whose supported property keys 
//             are being requested
// 
//
// Return Value:
//  Status
// 
//
/////////////////////////////////////////////////////////////////////////
HRESULT CSensorDDI:: OnSetProperties(
        __in IWDFFile* appId,
        __in LPWSTR ObjectID,
        __in IPortableDeviceValues* pValues,
        __out IPortableDeviceValues** ppResults
        )
{
    UNREFERENCED_PARAMETER(appId);

    HRESULT hr = S_OK;

    if ((ObjectID == NULL) ||
        (pValues       == NULL) ||
        (ppResults     == NULL))
    {
        hr = E_INVALIDARG;
        return hr;
    }

    return hr;
}

/////////////////////////////////////////////////////////////////////////
//
// CSensorDDI::OnClientConnect
//
// Routine Description:
//
//  This method is called by Sensor Class Extension when a client app connects
//  to a Sensor
//  
// Arguments:
//
//  ObjectID - String that represents the object whose supported property keys 
//             are being requested
// 
//
// Return Value:
//  Status
// 
//
/////////////////////////////////////////////////////////////////////////
HRESULT CSensorDDI:: OnClientConnect(
        __in IWDFFile* pClientFile,
        __in LPWSTR pwszObjectID
        )
{
    UNREFERENCED_PARAMETER(pClientFile);
    UNREFERENCED_PARAMETER(pwszObjectID);

    HRESULT hr = S_OK;

    m_dwClients++;

    if(1 == m_dwClients)
    {
        // Change the current report interval to the default
        
        CComPtr<IPortableDeviceValues> pReportInterval = NULL;
        hr = CoCreateInstance(CLSID_PortableDeviceValues,
                              NULL,
                              CLSCTX_INPROC_SERVER,
                              IID_IPortableDeviceValues,
                              (VOID**)&pReportInterval);
		if(SUCCEEDED(hr))
        {
	        pReportInterval.Release();
        }
    }

    return hr;
}

/////////////////////////////////////////////////////////////////////////
//
// CSensorDDI::OnClientDisconnect
//
// Routine Description:
//
//  This method is called by Sensor Class Extension when a client app disconnects
//  from a Sensor
//  
// Arguments:
//
//  ObjectID - String that represents the object whose supported property keys 
//             are being requested
// 
//
// Return Value:
//  Status
// 
//
/////////////////////////////////////////////////////////////////////////
HRESULT CSensorDDI:: OnClientDisconnect(
        __in IWDFFile* pClientFile,
        __in LPWSTR pwszObjectID
        )
{
    UNREFERENCED_PARAMETER(pClientFile);
    UNREFERENCED_PARAMETER(pwszObjectID);

    HRESULT hr = S_OK;

    m_dwClients--;

    if(0 == m_dwClients)
    {
        // Change the current report interval to zero
        
        CComPtr<IPortableDeviceValues> pReportInterval = NULL;
        hr = CoCreateInstance(CLSID_PortableDeviceValues,
                              NULL,
                              CLSCTX_INPROC_SERVER,
                              IID_IPortableDeviceValues,
                              (VOID**)&pReportInterval);
        if(SUCCEEDED(hr))
        {
	        pReportInterval.Release();
        }

    }

    return hr;
}

/////////////////////////////////////////////////////////////////////////
//
// CSensorDDI::OnClientSubscribeToEvents
//
// Routine Description:
//
//  This method is called by Sensor Class Extension when a client subscribes to
//  events by calling SetEventSink
//  
// Arguments:
//
//  ObjectID - String that represents the object whose supported property keys 
//             are being requested
// 
//
// Return Value:
//  Status
// 
//
/////////////////////////////////////////////////////////////////////////
HRESULT CSensorDDI:: OnClientSubscribeToEvents(
        __in IWDFFile* pClientFile,
        __in LPWSTR pwszObjectID
        )
{
    UNREFERENCED_PARAMETER(pClientFile);
    UNREFERENCED_PARAMETER(pwszObjectID);

    HRESULT hr = S_OK;
   
    hr = m_properties.OnClientSubscribeToEvents(pwszObjectID);
 
    return hr;
}

/////////////////////////////////////////////////////////////////////////
//
// CSensorDDI::OnClientUnsubscribeFromEvents
//
// Routine Description:
//
//  This method is called by Sensor Class Extension when a client unsubscribes from
//  events by calling SetEventSink(NULL)
//  
// Arguments:
//
//  ObjectID - String that represents the object whose supported property keys 
//             are being requested
// 
//
// Return Value:
//  Status
// 
//
/////////////////////////////////////////////////////////////////////////
HRESULT CSensorDDI:: OnClientUnsubscribeFromEvents(
        __in IWDFFile* pClientFile,
        __in LPWSTR pwszObjectID
        )
{
    UNREFERENCED_PARAMETER(pClientFile);
    UNREFERENCED_PARAMETER(pwszObjectID);

    HRESULT hr = S_OK;

    hr = m_properties.OnClientUnsubscribeToEvents(pwszObjectID);
 
    return hr;
}

HRESULT CSensorDDI:: OnProcessWpdMessage(
        __in IUnknown* pPortableDeviceValuesParamsUnknown,
        __in IUnknown* pPortableDeviceValuesResultsUnknown
        )
{
    UNREFERENCED_PARAMETER(pPortableDeviceValuesParamsUnknown);
    UNREFERENCED_PARAMETER(pPortableDeviceValuesResultsUnknown);

    HRESULT hr = S_OK;

    return hr;
}

//Asynchronous read call back function
void CSensorDDI:: OnAsyncReadCallback( BYTE* buffer, ULONG length )
{
    m_properties.ProcessAsynchRead( buffer ,length );

    return;
}