 /*

// THIS CODE AND INFORMATION IS PROVIDED "AS IS" WITHOUT WARRANTY OF
// ANY KIND, EITHER EXPRESSED OR IMPLIED, INCLUDING BUT NOT LIMITED TO
// THE IMPLIED WARRANTIES OF MERCHANTABILITY AND/OR FITNESS FOR A
// PARTICULAR PURPOSE.
//
// Copyright (c) Microsoft Corporation. All rights reserved


 Module:
      BadgeSensorDevice.cpp

 Description:
      Implements the CBadgeSensorDevice container class

--*/

#include "internal.h"
#include "Defines.h"
#include "SensorDriver.h"
#include "SensorDevice.h"

#include "SensorDevice.tmh"

const PROPERTYKEY g_SupportedAccelerometerProperties[] =
{
    SENSOR_PROPERTY_TYPE,
    SENSOR_PROPERTY_STATE,
    SENSOR_PROPERTY_MIN_REPORT_INTERVAL,
    SENSOR_PROPERTY_CURRENT_REPORT_INTERVAL,
    SENSOR_PROPERTY_PERSISTENT_UNIQUE_ID,
    SENSOR_PROPERTY_MANUFACTURER, 
    SENSOR_PROPERTY_MODEL,
    SENSOR_PROPERTY_SERIAL_NUMBER,
    SENSOR_PROPERTY_FRIENDLY_NAME,
    SENSOR_PROPERTY_DESCRIPTION, 
    SENSOR_PROPERTY_CONNECTION_TYPE,
    SENSOR_PROPERTY_CHANGE_SENSITIVITY,
    WPD_FUNCTIONAL_OBJECT_CATEGORY,
};

const PROPERTYKEY g_SupportedAccelerometerDataFields[] =
{
	SENSOR_DATA_TYPE_TIMESTAMP,
	SENSOR_DATA_TYPE_ACCELERATION_X_G,
	SENSOR_DATA_TYPE_ACCELERATION_Y_G,
	SENSOR_DATA_TYPE_ACCELERATION_Z_G,
};

const PROPERTYKEY g_SupportedAmbientLightProperties[] =
{
    SENSOR_PROPERTY_TYPE,
    SENSOR_PROPERTY_STATE,
    SENSOR_PROPERTY_MIN_REPORT_INTERVAL,
    SENSOR_PROPERTY_CURRENT_REPORT_INTERVAL,
    SENSOR_PROPERTY_PERSISTENT_UNIQUE_ID,
    SENSOR_PROPERTY_MANUFACTURER, 
    SENSOR_PROPERTY_MODEL,
    SENSOR_PROPERTY_SERIAL_NUMBER,
    SENSOR_PROPERTY_FRIENDLY_NAME,
    SENSOR_PROPERTY_DESCRIPTION,
    SENSOR_PROPERTY_CONNECTION_TYPE,        
    SENSOR_PROPERTY_LIGHT_RESPONSE_CURVE,
    SENSOR_PROPERTY_CHANGE_SENSITIVITY,
    WPD_FUNCTIONAL_OBJECT_CATEGORY,
};

const PROPERTYKEY g_SupportedAmbientLightDataFields[] =
{
	SENSOR_DATA_TYPE_TIMESTAMP,
	SENSOR_DATA_TYPE_LIGHT_LEVEL_LUX,
};

const PROPERTYKEY g_SupportedSwitchArrayProperties[] =
{
    SENSOR_PROPERTY_TYPE,
    SENSOR_PROPERTY_STATE,
    SENSOR_PROPERTY_MIN_REPORT_INTERVAL,
    SENSOR_PROPERTY_CURRENT_REPORT_INTERVAL,
    SENSOR_PROPERTY_PERSISTENT_UNIQUE_ID,
    SENSOR_PROPERTY_MANUFACTURER, 
    SENSOR_PROPERTY_MODEL,
    SENSOR_PROPERTY_SERIAL_NUMBER,
    SENSOR_PROPERTY_FRIENDLY_NAME,
    SENSOR_PROPERTY_DESCRIPTION, 
    SENSOR_PROPERTY_CONNECTION_TYPE,
    SENSOR_PROPERTY_CHANGE_SENSITIVITY,
    WPD_FUNCTIONAL_OBJECT_CATEGORY,
};

const PROPERTYKEY g_SupportedSwitchArrayDataFields[] =
{
	SENSOR_DATA_TYPE_TIMESTAMP,
	SENSOR_DATA_TYPE_BOOLEAN_SWITCH_ARRAY_STATE,
};

/////////////////////////////////////////////////////////////////////////
//
// CBadgeSensorDevice::CBadgeSensorDevice
//
// Object constructor function
//
/////////////////////////////////////////////////////////////////////////
CBadgeSensorDevice::CBadgeSensorDevice() :
    m_pSensorProperties(NULL),
    m_pSensorDataFields(NULL),
    m_fBadgeSensorDeviceInitialized(FALSE),
    m_dwEventSubscribers(0),
    m_fValidDataEvent(FALSE)
{

}

/////////////////////////////////////////////////////////////////////////
//
// CBadgeSensorDevice::~CBadgeSensorDevice
//
// Object destructor function
//
/////////////////////////////////////////////////////////////////////////
CBadgeSensorDevice::~CBadgeSensorDevice()
{
    Uninitialize();
}

/////////////////////////////////////////////////////////////////////////
//
// CBadgeSensorDevice::Initialize
//
// Initializes the PROPERTYKEY/PROPVARIANT values for
// the Ambient Light Sensor Supported Properties & Supported Data
//
/////////////////////////////////////////////////////////////////////////
HRESULT CBadgeSensorDevice::Initialize(SensorType sensType)
{
    CComCritSecLock<CComAutoCriticalSection> scopeLock(m_CriticalSection); // Make this call thread safe

    // Check if we are already initialized
    HRESULT hr = (TRUE == IsInitialized()) ? E_UNEXPECTED : S_OK;

    Trace(TRACE_LEVEL_INFORMATION, "%!FUNC! Entry");

    if(SUCCEEDED(hr))
    {
        m_SensorType = sensType;

        if(NULL == m_pSupportedSensorProperties)
        {
            // Create a new PortableDeviceKeyCollection to store the supported property KEYS
            hr = CoCreateInstance(CLSID_PortableDeviceKeyCollection,
                                  NULL,
                                  CLSCTX_INPROC_SERVER,
                                  IID_IPortableDeviceKeyCollection,
                                  (VOID**)&m_pSupportedSensorProperties);
        }

        if(NULL == m_pSensorProperties)
        {
            // Create a new PortableDeviceValues to store the property VALUES
            hr = CoCreateInstance(CLSID_PortableDeviceValues,
                                  NULL,
                                  CLSCTX_INPROC_SERVER,
                                  IID_IPortableDeviceValues,
                                  (VOID**)&m_pSensorProperties);
        }
        
        if(NULL == m_pSupportedSensorDataFields)
        {
            // Create a new PortableDeviceValues to store the supported datafield KEYS
            hr = CoCreateInstance(CLSID_PortableDeviceKeyCollection,
                                  NULL,
                                  CLSCTX_INPROC_SERVER,
                                  IID_IPortableDeviceKeyCollection,
                                  (VOID**)&m_pSupportedSensorDataFields);
        }

        if(NULL == m_pSensorDataFields)
        {
            // Create a new PortableDeviceValues to store the datafield VALUES
            hr = CoCreateInstance(CLSID_PortableDeviceValues,
                                  NULL,
                                  CLSCTX_INPROC_SERVER,
                                  IID_IPortableDeviceValues,
                                  (VOID**)&m_pSensorDataFields);
        }

        if(NULL == m_pSettableSensorProperties)
        {
            // Create a new PortableDeviceValues to store the settable property keys
            hr = CoCreateInstance(CLSID_PortableDeviceKeyCollection,
                                  NULL,
                                  CLSCTX_INPROC_SERVER,
                                  IID_IPortableDeviceKeyCollection,
                                  (VOID**)&m_pSettableSensorProperties);
        }

        //Configure the "settable properties" for the sensor
        // All Sensors have the same settable properties
        if(SUCCEEDED(hr))
        {        
            hr = SetSettableSensorProperties();
        }

        if(SUCCEEDED(hr))
        {
            switch(sensType)
            {
            case Accelerometer:
                hr = InitializeAccelerometer();
                break;
            case AmbientLight:
                hr = InitializeAmbientLight();
                break;
            case SwitchArrayLeft:
                hr = InitializeSwitchArrayLeft();
                break;
            case SwitchArrayRight:
                hr = InitializeSwitchArrayRight();
                break;
            default:
                break;
            }
        }

        if(SUCCEEDED(hr))
        {
            m_fBadgeSensorDeviceInitialized = TRUE;
        }
    }

    return hr;
}

HRESULT CBadgeSensorDevice::InitializeAmbientLight(VOID)
{
    HRESULT hr = S_OK;

    hr = AddAmbientLightPropertyKeys();
    
    if(SUCCEEDED(hr))
    {
        hr = AddAmbientLightDataFieldKeys();
    }

    if(SUCCEEDED(hr))
    {
        hr = SetAmbientLightDefaultValues();
    }

    return hr;
}

HRESULT CBadgeSensorDevice::InitializeAccelerometer(VOID)
{
    HRESULT hr = S_OK;

    hr = AddAccelerometerPropertyKeys();

    if(SUCCEEDED(hr))
    {
        hr = AddAccelerometerDataFieldKeys();
    }

    if(SUCCEEDED(hr))
    {
        hr = SetAccelerometerDefaultValues();
    }

    return hr;
}

HRESULT CBadgeSensorDevice::InitializeSwitchArrayLeft(VOID)
{
    HRESULT hr = S_OK;

    hr = AddSwitchArrayPropertyKeys();

    if(SUCCEEDED(hr))
    {
        hr = AddSwitchArrayDataFieldKeys();
    }

    if(SUCCEEDED(hr))
    {
        hr = SetSwitchArrayLeftDefaultValues();
    }

    return hr;
}


HRESULT CBadgeSensorDevice::InitializeSwitchArrayRight(VOID)
{
    HRESULT hr = S_OK;

    hr = AddSwitchArrayPropertyKeys();

    if(SUCCEEDED(hr))
    {
        hr = AddSwitchArrayDataFieldKeys();
    }

    if(SUCCEEDED(hr))
    {
        hr = SetSwitchArrayRightDefaultValues();
    }

    return hr;
}


/////////////////////////////////////////////////////////////////////////
//
// CBadgeSensorDevice::Uninitialize
//
// Initializes the PROPERTYKEY/PROPVARIANT values for
// the Ambient Light Sensor Supported Properties & Supported Data
//
/////////////////////////////////////////////////////////////////////////
HRESULT CBadgeSensorDevice::Uninitialize()
{
    CComCritSecLock<CComAutoCriticalSection> scopeLock(m_CriticalSection); // Make this call thread safe

    // Check if we are already initialized
    HRESULT hr = (FALSE == IsInitialized()) ? E_UNEXPECTED : S_OK;

    if(SUCCEEDED(hr))
    {
        hr = m_pSupportedSensorProperties->Clear();

        if(SUCCEEDED(hr))
            hr = m_pSupportedSensorDataFields->Clear();

        if(SUCCEEDED(hr))
            hr = m_pSensorProperties->Clear();

        if(SUCCEEDED(hr))
            hr = m_pSensorDataFields->Clear();

        m_pSensorProperties = NULL;
        m_pSensorDataFields = NULL;
        m_pSupportedSensorProperties = NULL;
        m_pSupportedSensorDataFields = NULL;

        m_fBadgeSensorDeviceInitialized = FALSE;
    }

    return hr;
}
///////////////////////////////////////////////////////////////////////////
//
// CBadgeSensorDevice::GetSettableProperties
//
// Gets the list of SettableProperties for the ALS Device
// 
// Returns a collection of PROPERTYKEYS
//
///////////////////////////////////////////////////////////////////////////
HRESULT CBadgeSensorDevice::GetSettableProperties(IPortableDeviceKeyCollection **ppKeys)
{
    CComCritSecLock<CComAutoCriticalSection> scopeLock(m_CriticalSection); // Make this call thread safe

    HRESULT hr = S_OK;

    if(NULL == ppKeys)
    {
        hr = E_INVALIDARG;
    }

    // Make sure we have been Initialized
    if((FALSE == IsInitialized()) || (NULL == m_pSupportedSensorProperties))
    {
        hr = E_UNEXPECTED;
    }
    
    if(SUCCEEDED(hr))
    {
        hr = m_pSettableSensorProperties.CopyTo(ppKeys);
    }

    return hr;
}


///////////////////////////////////////////////////////////////////////////
//
// CBadgeSensorDevice::GetSupportedProperties
//
// Gets the list of SupportedProperties for the ALS Device
// 
// Returns a collection of PROPERTYKEYS
//
///////////////////////////////////////////////////////////////////////////
HRESULT CBadgeSensorDevice::GetSupportedProperties(IPortableDeviceKeyCollection **ppKeys)
{
    CComCritSecLock<CComAutoCriticalSection> scopeLock(m_CriticalSection); // Make this call thread safe

    HRESULT hr = S_OK;

    if(NULL == ppKeys)
    {
        hr = E_INVALIDARG;
    }

    // Make sure we have been Initialized
    if((FALSE == IsInitialized()) || (NULL == m_pSupportedSensorProperties))
    {
        hr = E_UNEXPECTED;
    }
    
    if(SUCCEEDED(hr))
    {
        hr = m_pSupportedSensorProperties.CopyTo(ppKeys);
    }

    return hr;
}

//////////////////////////////////////////////////////////////////////////////////////
//
// CBadgeSensorDevice::GetSupportedDataFields
//
// Gets the list of SupportedDataFields for the ALS Device
// 
// Returns a collection of PROPERTYKEYS
//
//////////////////////////////////////////////////////////////////////////////////////
HRESULT CBadgeSensorDevice::GetSupportedDataFields(IPortableDeviceKeyCollection **ppKeys)
{
    CComCritSecLock<CComAutoCriticalSection> scopeLock(m_CriticalSection); // Make this call thread safe

    HRESULT hr = S_OK;

    if(NULL == ppKeys)
    {
        hr = E_INVALIDARG;
    }

    // Make sure we have been Initialized
    if((FALSE == IsInitialized()) || (NULL == m_pSupportedSensorDataFields))
    {
        hr = E_UNEXPECTED;
    }
    
    if(SUCCEEDED(hr))
    {
        hr = m_pSupportedSensorDataFields.CopyTo(ppKeys);
    }
    
    return hr;
}

///////////////////////////////////////////////////////////////////////////
//
// CBadgeSensorDevice::GetProperty
//
// Gets the Property Value for a given Property key.
// 
// Called to retrive a property of the ALS Sensor Device
//
///////////////////////////////////////////////////////////////////////////
HRESULT CBadgeSensorDevice::GetProperty(REFPROPERTYKEY key, PROPVARIANT *pValue)
{
    CComCritSecLock<CComAutoCriticalSection> scopeLock(m_CriticalSection); // Make this call thread safe

    HRESULT hr = S_OK;

    if(NULL == m_pSensorProperties)
    {
        hr = E_UNEXPECTED;
    }
    else
    {
        // Retrieve the value
        hr = m_pSensorProperties->GetValue(key, pValue);
    }

    return hr;
}

///////////////////////////////////////////////////////////////////////////
//
// CBadgeSensorDevice::GetDataField
//
// Gets the Data Field Value for a given Property key.
// 
// Called to retrieve a data field of the ALS Sensor Device
//
///////////////////////////////////////////////////////////////////////////
HRESULT CBadgeSensorDevice::GetDataField(REFPROPERTYKEY key, PROPVARIANT *pValue)
{ 
    CComCritSecLock<CComAutoCriticalSection> scopeLock(m_CriticalSection); // Make this call thread safe

    HRESULT hr = S_OK;

    if(NULL == m_pSensorDataFields)
    {
        hr = E_UNEXPECTED;
    }
    else
    {
        // Retrieve the value
        hr = m_pSensorDataFields->GetValue(key, pValue);
    }

    return hr;
}

///////////////////////////////////////////////////////////////////////////
//
// CBadgeSensorDevice::GetAllDataFieldValues
//
// Gets the all the Data  values 
// 
// Called to retrieve all data field values of the ALS Sensor Device
//
///////////////////////////////////////////////////////////////////////////
HRESULT CBadgeSensorDevice::GetAllDataFieldValues(IPortableDeviceValues* pValues)
{
    CComCritSecLock<CComAutoCriticalSection> scopeLock(m_CriticalSection); // Make this call thread safe

    HRESULT hr = S_OK;

    if(NULL == m_pSensorDataFields)
    {
        hr = E_UNEXPECTED;
    }
    else
    {
        DWORD dwCount = 0;
        hr = m_pSensorDataFields->GetCount( &dwCount );

        if( SUCCEEDED(hr) )
        {
            for( DWORD i = 0 ; i < dwCount ; i++ )
            {
                PROPERTYKEY key;
                PROPVARIANT var;
                PropVariantInit( &var );
                hr = m_pSensorDataFields->GetAt( i, &key, &var );
                
                if( SUCCEEDED(hr) )
                {
                    hr = pValues->SetValue( key, &var );
                }

                PropVariantClear(&var);
            }
        }

    }

    return hr;
}


///////////////////////////////////////////////////////////////////////////
//
// CBadgeSensorDevice::SetProperty
//
// Sets the Property Value for a given Property key.
// 
// Called during Initialize and when a property is read from the ALS Device
//
///////////////////////////////////////////////////////////////////////////
HRESULT CBadgeSensorDevice::SetProperty(REFPROPERTYKEY key, const PROPVARIANT *pValue)
{
    CComCritSecLock<CComAutoCriticalSection> scopeLock(m_CriticalSection); // Make this call thread safe
    
    HRESULT hr = (NULL == pValue) ? E_INVALIDARG : S_OK;

    if(SUCCEEDED(hr))
    {
        if (TRUE == IsEqualPropertyKey(key, SENSOR_PROPERTY_CURRENT_REPORT_INTERVAL)
            && (VT_EMPTY != V_VT(pValue)))
        {
            hr = E_INVALIDARG;
      
            // Validate Range for Report Interval 
            if (VT_UI4 == V_VT(pValue))
            {
                ULONG ulReportInterval = V_UI4(pValue);
                if (ulReportInterval >= DEFAULT_MIN_REPORT_INTERVAL)
                {
                    hr = S_OK;
                }
            }
        }
        else if (TRUE == IsEqualPropertyKey(key, SENSOR_PROPERTY_CHANGE_SENSITIVITY)
                 && (VT_EMPTY != V_VT(pValue)))
        {
            hr = E_INVALIDARG;
      
            // Validate Range for Report Interval 
            if (VT_R4 == V_VT(pValue))
            {
                FLOAT ulSensitivity = V_R4(pValue);
                if (ulSensitivity >= MIN_CHANGE_SENSITIVITY && 
                    ulSensitivity <= MAX_CHANGE_SENSITIVITY )
                {
                    hr = S_OK;
                }
            }
        }

        if(SUCCEEDED(hr))
        {
            hr = m_pSensorProperties->SetValue(key, pValue);
        }
    }

    return hr;
}

///////////////////////////////////////////////////////////////////////////
//
// CBadgeSensorDevice::SetDataField
//
// Sets the Data Field Value for a given DataField key.
// 
// Called during Initialize and when data is read from the ALS Device
//
///////////////////////////////////////////////////////////////////////////
HRESULT CBadgeSensorDevice::SetDataField(REFPROPERTYKEY key, const PROPVARIANT *pValue)
{
    CComCritSecLock<CComAutoCriticalSection> scopeLock(m_CriticalSection); // Make this call thread safe

    HRESULT hr = S_OK;

    if(NULL == pValue)
    {
        hr = E_INVALIDARG;
    }

    if(NULL == m_pSensorDataFields)
    {
        hr = E_POINTER;
    }
    
    if(SUCCEEDED(hr))
    {
        hr = m_pSensorDataFields->SetValue(key, pValue);
    }

    return hr;
}

///////////////////////////////////////////////////////////////////////////
//
// CBadgeSensorDevice::Subscribe
//
// Sets the status of event subscribers
// 
// 
///////////////////////////////////////////////////////////////////////////
void CBadgeSensorDevice::Subscribe()
{
    CComCritSecLock<CComAutoCriticalSection> scopeLock(m_CriticalSection);

    m_dwEventSubscribers++;

    return;
}


///////////////////////////////////////////////////////////////////////////
//
// CBadgeSensorDevice::Unsubscribe
//
// Sets the status of event subscribers
// 
// 
///////////////////////////////////////////////////////////////////////////
void CBadgeSensorDevice::Unsubscribe()
{
    CComCritSecLock<CComAutoCriticalSection> scopeLock(m_CriticalSection);

    m_dwEventSubscribers--;

    return;
}

///////////////////////////////////////////////////////////////////////////
//
// CBadgeSensorDevice::GetSubscriberCount
//
// Retrieves the status of event subscribers
// 
// 
///////////////////////////////////////////////////////////////////////////
DWORD CBadgeSensorDevice::GetSubscriberCount(void)
{
    CComCritSecLock<CComAutoCriticalSection> scopeLock(m_CriticalSection);

    return m_dwEventSubscribers;
}

/////////////////////////////////////////////////////////////////////////
//
// CBadgeSensorDevice::AddAmbientLightPropertyKeys
//
// Copies the PROPERTYKEYS for Ambient Light Sensor Supported Properties 
// and sets the Values to VT_EMPTY
//
/////////////////////////////////////////////////////////////////////////
HRESULT CBadgeSensorDevice::AddAmbientLightPropertyKeys()
{
    HRESULT hr = S_OK;

    for (DWORD dwIndex = 0; dwIndex < ARRAY_SIZE(g_SupportedAmbientLightProperties); dwIndex++)
    {
        PROPVARIANT var;
        PropVariantInit(&var);

        // Initialize all the PROPERTYKEY values to VT_EMPTY
        hr = SetProperty(g_SupportedAmbientLightProperties[dwIndex], &var);

        // Also add the PROPERTYKEY to the list of supported properties
        if(SUCCEEDED(hr))
        {
            hr = m_pSupportedSensorProperties->Add(g_SupportedAmbientLightProperties[dwIndex]);
        }

        PropVariantClear(&var);
    }

    return hr;
}

/////////////////////////////////////////////////////////////////////////
//
// CBadgeSensorDevice::AddAmbientLightDataFieldKeys
//
// Copies the PROPERTYKEYS for Ambient Light Sensor Supported DataFields 
// and sets the Values to VT_EMPTY
//
/////////////////////////////////////////////////////////////////////////
HRESULT CBadgeSensorDevice::AddAmbientLightDataFieldKeys()
{
    HRESULT hr = S_OK;

    for (DWORD dwIndex = 0; dwIndex < ARRAY_SIZE(g_SupportedAmbientLightDataFields); dwIndex++)
    {
        PROPVARIANT var;
        PropVariantInit(&var);

        // Initialize all the PROPERTYKEY values to VT_EMPTY
        hr = SetDataField(g_SupportedAmbientLightDataFields[dwIndex], &var);

        // Also add the PROPERTYKEY to the list of supported data fields
        if(SUCCEEDED(hr))
        {
            hr = m_pSupportedSensorDataFields->Add(g_SupportedAmbientLightDataFields[dwIndex]);
        }

        PropVariantClear(&var);
    }

    return hr;
}


/////////////////////////////////////////////////////////////////////////
//
// CBadgeSensorDevice::AddAccelerometerPropertyKeys
//
// Copies the PROPERTYKEYS for Ambient Light Sensor Supported Properties 
// and sets the Values to VT_EMPTY
//
/////////////////////////////////////////////////////////////////////////
HRESULT CBadgeSensorDevice::AddAccelerometerPropertyKeys()
{
    HRESULT hr = S_OK;

    for (DWORD dwIndex = 0; dwIndex < ARRAY_SIZE(g_SupportedAccelerometerProperties); dwIndex++)
    {
        PROPVARIANT var;
        PropVariantInit(&var);

        // Initialize all the PROPERTYKEY values to VT_EMPTY
        hr = SetProperty(g_SupportedAccelerometerProperties[dwIndex], &var);

        // Also add the PROPERTYKEY to the list of supported properties
        if(SUCCEEDED(hr))
        {
            hr = m_pSupportedSensorProperties->Add(g_SupportedAccelerometerProperties[dwIndex]);
        }

        PropVariantClear(&var);
    }

    return hr;
}

/////////////////////////////////////////////////////////////////////////
//
// CBadgeSensorDevice::AddAccelerometerDataFieldKeys
//
// Copies the PROPERTYKEYS for Ambient Light Sensor Supported DataFields 
// and sets the Values to VT_EMPTY
//
/////////////////////////////////////////////////////////////////////////
HRESULT CBadgeSensorDevice::AddAccelerometerDataFieldKeys()
{
    HRESULT hr = S_OK;

    for (DWORD dwIndex = 0; dwIndex < ARRAY_SIZE(g_SupportedAccelerometerDataFields); dwIndex++)
    {
        PROPVARIANT var;
        PropVariantInit(&var);

        // Initialize all the PROPERTYKEY values to VT_EMPTY
        hr = SetDataField(g_SupportedAccelerometerDataFields[dwIndex], &var);

        // Also add the PROPERTYKEY to the list of supported data fields
        if(SUCCEEDED(hr))
        {
            hr = m_pSupportedSensorDataFields->Add(g_SupportedAccelerometerDataFields[dwIndex]);
        }

        PropVariantClear(&var);
    }

    return hr;
}


/////////////////////////////////////////////////////////////////////////
//
// CBadgeSensorDevice::AddSwitchArrayPropertyKeys
//
// Copies the PROPERTYKEYS for Ambient Light Sensor Supported Properties 
// and sets the Values to VT_EMPTY
//
/////////////////////////////////////////////////////////////////////////
HRESULT CBadgeSensorDevice::AddSwitchArrayPropertyKeys()
{
    HRESULT hr = S_OK;

    for (DWORD dwIndex = 0; dwIndex < ARRAY_SIZE(g_SupportedSwitchArrayProperties); dwIndex++)
    {
        PROPVARIANT var;
        PropVariantInit(&var);

        // Initialize all the PROPERTYKEY values to VT_EMPTY
        hr = SetProperty(g_SupportedSwitchArrayProperties[dwIndex], &var);

        // Also add the PROPERTYKEY to the list of supported properties
        if(SUCCEEDED(hr))
        {
            hr = m_pSupportedSensorProperties->Add(g_SupportedSwitchArrayProperties[dwIndex]);
        }

        PropVariantClear(&var);
    }

    return hr;
}

/////////////////////////////////////////////////////////////////////////
//
// CBadgeSensorDevice::AddSwitchArrayDataFieldKeys
//
// Copies the PROPERTYKEYS for Ambient Light Sensor Supported DataFields 
// and sets the Values to VT_EMPTY
//
/////////////////////////////////////////////////////////////////////////
HRESULT CBadgeSensorDevice::AddSwitchArrayDataFieldKeys()
{
    HRESULT hr = S_OK;

    for (DWORD dwIndex = 0; dwIndex < ARRAY_SIZE(g_SupportedSwitchArrayDataFields); dwIndex++)
    {
        PROPVARIANT var;
        PropVariantInit(&var);

        // Initialize all the PROPERTYKEY values to VT_EMPTY
        hr = SetDataField(g_SupportedSwitchArrayDataFields[dwIndex], &var);

        // Also add the PROPERTYKEY to the list of supported data fields
        if(SUCCEEDED(hr))
        {
            hr = m_pSupportedSensorDataFields->Add(g_SupportedSwitchArrayDataFields[dwIndex]);
        }

        PropVariantClear(&var);
    }

    return hr;
}

/////////////////////////////////////////////////////////////////////////
//
// CBadgeSensorDevice::SetSettableSensorProperties
//
// Fills in the settable property keys for the device
//
//
//
/////////////////////////////////////////////////////////////////////////
HRESULT CBadgeSensorDevice::SetSettableSensorProperties(VOID)
{
    HRESULT hr = S_OK;

    if(NULL == m_pSettableSensorProperties)
    {
        hr = E_POINTER;
    }

    if(SUCCEEDED(hr))
        hr = m_pSettableSensorProperties->Add(SENSOR_PROPERTY_CHANGE_SENSITIVITY);

    if(SUCCEEDED(hr))
        hr = m_pSettableSensorProperties->Add(SENSOR_PROPERTY_CURRENT_REPORT_INTERVAL);

    return hr;
}

/////////////////////////////////////////////////////////////////////////
//
// CBadgeSensorDevice::RemoveProperty
//
// Removes the PROPERTYKEY passed in from the list of Supported Properties
//
//
//
/////////////////////////////////////////////////////////////////////////
HRESULT CBadgeSensorDevice::RemoveProperty(REFPROPERTYKEY key)
{
    CComCritSecLock<CComAutoCriticalSection> scopeLock(m_CriticalSection);

    HRESULT hr = S_OK;
    DWORD cKeys = 0;
    PROPERTYKEY tempKey;

    if (NULL != m_pSupportedSensorProperties)
    {
        hr = m_pSupportedSensorProperties->GetCount(&cKeys);

        if (SUCCEEDED(hr))
        {
            for (DWORD dwIndex = 0; dwIndex < cKeys; ++dwIndex)
            {
                hr = m_pSupportedSensorProperties->GetAt(dwIndex, &tempKey);

                if (SUCCEEDED(hr))
                {
                    if (IsEqualPropertyKey(tempKey, key))
                    {
                        hr = m_pSupportedSensorProperties->RemoveAt(dwIndex);

                        if (SUCCEEDED(hr))
                            hr = m_pSensorProperties->RemoveValue(key); 

                        break;
                    }
                }
            }
        }
    }

    return hr;
}

/////////////////////////////////////////////////////////////////////////
//
// CBadgeSensorDevice::RemoveDataField
//
// Removes the PROPERTYKEY passed in from the list of Supported DataFields
//
//
//
/////////////////////////////////////////////////////////////////////////
HRESULT CBadgeSensorDevice::RemoveDataField(REFPROPERTYKEY key)
{
    CComCritSecLock<CComAutoCriticalSection> scopeLock(m_CriticalSection);

    HRESULT hr = S_OK;
    DWORD cKeys = 0;
    PROPERTYKEY tempKey;

    if (NULL != m_pSupportedSensorDataFields)
    {
        hr = m_pSupportedSensorDataFields->GetCount(&cKeys);

        if (SUCCEEDED(hr))
        {
            for (DWORD dwIndex = 0; dwIndex < cKeys; ++dwIndex)
            {
                hr = m_pSupportedSensorDataFields->GetAt(dwIndex, &tempKey);

                if (SUCCEEDED(hr))
                {
                    if (IsEqualPropertyKey(tempKey, key))
                    {
                        hr = m_pSupportedSensorDataFields->RemoveAt(dwIndex);

                        if (SUCCEEDED(hr))
                            hr = m_pSensorDataFields->RemoveValue(key); 

                        break;
                    }
                }
            }
        }
    }

    return hr;
}



/////////////////////////////////////////////////////////////////////////
//
// CBadgeSensorDevice::SetAccelerometerDefaultValues
//
// Fills in default values for most Ambient Light Sensor Properties and 
// Data Fields.
//
//
/////////////////////////////////////////////////////////////////////////
HRESULT CBadgeSensorDevice::SetAccelerometerDefaultValues()
{
    HRESULT hr = S_OK;

    if((NULL == m_pSensorProperties) || (NULL == m_pSensorDataFields))
    {
        hr = E_POINTER;
    }

    // *****************************************************************************************
    // Default values for SENSOR PROPERTIES
    // *****************************************************************************************

    if(SUCCEEDED(hr))
        hr = m_pSensorProperties->SetGuidValue(WPD_FUNCTIONAL_OBJECT_CATEGORY, SENSOR_CATEGORY_MOTION);

    if(SUCCEEDED(hr))
        hr = m_pSensorProperties->SetGuidValue(SENSOR_PROPERTY_TYPE, SENSOR_TYPE_ACCELEROMETER_3D);

    if(SUCCEEDED(hr))
        hr = m_pSensorProperties->SetUnsignedIntegerValue(SENSOR_PROPERTY_STATE, SENSOR_STATE_MIN);

    if(SUCCEEDED(hr))
        hr = m_pSensorProperties->SetUnsignedIntegerValue(SENSOR_PROPERTY_MIN_REPORT_INTERVAL, DEFAULT_MIN_REPORT_INTERVAL);

    if(SUCCEEDED(hr))
        hr = m_pSensorProperties->SetUnsignedIntegerValue(SENSOR_PROPERTY_CURRENT_REPORT_INTERVAL, AWAKE_CURRENT_REPORT_INTERVAL);
    
    if(SUCCEEDED(hr))
        hr = m_pSensorProperties->SetStringValue(SENSOR_PROPERTY_MANUFACTURER, DEVICE_MANUFACTURER_VALUE);

    if(SUCCEEDED(hr))
        hr = m_pSensorProperties->SetStringValue(SENSOR_PROPERTY_MODEL, DEVICE_MODEL_VALUE);

    if(SUCCEEDED(hr))
        hr = m_pSensorProperties->SetStringValue(SENSOR_PROPERTY_SERIAL_NUMBER, DEVICE_SERIAL_NUMBER_VALUE);

    if(SUCCEEDED(hr))
        hr = m_pSensorProperties->SetStringValue(SENSOR_PROPERTY_FRIENDLY_NAME, SENSOR_ACCELEROMETER_FRIENDLY_NAME);

    if(SUCCEEDED(hr))
        hr = m_pSensorProperties->SetStringValue(SENSOR_PROPERTY_DESCRIPTION, SENSOR_ACCELEROMETER_FRIENDLY_NAME);

    if(SUCCEEDED(hr))
        hr = m_pSensorProperties->SetUnsignedIntegerValue(SENSOR_PROPERTY_CONNECTION_TYPE, SENSOR_CONNECTION_TYPE_PC_EXTERNAL);

    if(SUCCEEDED(hr))
        hr = m_pSensorProperties->SetUnsignedIntegerValue(SENSOR_PROPERTY_CHANGE_SENSITIVITY, DEFAULT_CHANGE_SENSITIVITY);
   
    // *****************************************************************************************
    // Default values for SENSOR DATA FIELDS
    // *****************************************************************************************
    
    if(SUCCEEDED(hr))
    {
        PROPVARIANT var;
        PropVariantInit( &var );

        //Get the current time in  SYSTEMTIME format
        SYSTEMTIME st;
        ::GetSystemTime(&st);

        // Convert the SYSTEMTIME into FILETIME
        FILETIME ft;
        if(FALSE == ::SystemTimeToFileTime(&st, &ft))
        {
            hr = HRESULT_FROM_WIN32(GetLastError());
        }

        if (S_OK == hr)
        {
            var.vt                      = VT_FILETIME;
            var.filetime.dwLowDateTime  = ft.dwLowDateTime;
            var.filetime.dwHighDateTime = ft.dwHighDateTime;
            hr = m_pSensorDataFields->SetValue(SENSOR_DATA_TYPE_TIMESTAMP, &var);
        }

        PropVariantClear( &var );
    }

    if(SUCCEEDED(hr))
        hr = m_pSensorDataFields->SetFloatValue(SENSOR_DATA_TYPE_ACCELERATION_X_G, DEFAULT_ACCELERATION_G);

    if(SUCCEEDED(hr))
        hr = m_pSensorDataFields->SetFloatValue(SENSOR_DATA_TYPE_ACCELERATION_Y_G, DEFAULT_ACCELERATION_G);

    if(SUCCEEDED(hr))
        hr = m_pSensorDataFields->SetFloatValue(SENSOR_DATA_TYPE_ACCELERATION_Z_G, DEFAULT_ACCELERATION_G);

    return hr;
}



HRESULT CBadgeSensorDevice::SetAmbientLightDefaultValues()
{
    HRESULT hr = S_OK;

    if((NULL == m_pSensorProperties) || (NULL == m_pSensorDataFields))
    {
        hr = E_POINTER;
    }

    // *****************************************************************************************
    // Default values for SENSOR PROPERTIES
    // *****************************************************************************************

    if(SUCCEEDED(hr))
        hr = m_pSensorProperties->SetGuidValue(WPD_FUNCTIONAL_OBJECT_CATEGORY, SENSOR_CATEGORY_LIGHT);

    if(SUCCEEDED(hr))
        hr = m_pSensorProperties->SetGuidValue(SENSOR_PROPERTY_TYPE, SENSOR_TYPE_AMBIENT_LIGHT);

    if(SUCCEEDED(hr))
        hr = m_pSensorProperties->SetUnsignedIntegerValue(SENSOR_PROPERTY_STATE, SENSOR_STATE_MIN);

    if(SUCCEEDED(hr))
        hr = m_pSensorProperties->SetUnsignedIntegerValue(SENSOR_PROPERTY_MIN_REPORT_INTERVAL, DEFAULT_MIN_REPORT_INTERVAL);

    if(SUCCEEDED(hr))
        hr = m_pSensorProperties->SetUnsignedIntegerValue(SENSOR_PROPERTY_CURRENT_REPORT_INTERVAL, AWAKE_CURRENT_REPORT_INTERVAL);

    if(SUCCEEDED(hr))
        hr = m_pSensorProperties->SetStringValue(SENSOR_PROPERTY_MANUFACTURER, DEVICE_MANUFACTURER_VALUE);

    if(SUCCEEDED(hr))
        hr = m_pSensorProperties->SetStringValue(SENSOR_PROPERTY_MODEL, DEVICE_MODEL_VALUE);

    if(SUCCEEDED(hr))
        hr = m_pSensorProperties->SetStringValue(SENSOR_PROPERTY_SERIAL_NUMBER, DEVICE_SERIAL_NUMBER_VALUE);

    if(SUCCEEDED(hr))
        hr = m_pSensorProperties->SetStringValue(SENSOR_PROPERTY_FRIENDLY_NAME, SENSOR_LIGHT_FRIENDLY_NAME);

    if(SUCCEEDED(hr))
        hr = m_pSensorProperties->SetStringValue(SENSOR_PROPERTY_DESCRIPTION, SENSOR_LIGHT_FRIENDLY_NAME);

    // ********************************************************************************************
    // Setting CONNECTION_TYPE to PC_INTEGRATED simulates an inbuilt ambient light sensor
    // This is useful for activating the adaptive brightness feature in Windows 7
    // ********************************************************************************************
    if(SUCCEEDED(hr))
        hr = m_pSensorProperties->SetUnsignedIntegerValue(SENSOR_PROPERTY_CONNECTION_TYPE, SENSOR_CONNECTION_TYPE_PC_INTEGRATED);

    // ********************************************************************************************
    // Setting LIGHT_CURVE to a default value to activate the adaptive brightness feature in
    // Windows 7. The actual value should be tailored to the specific hardware.
    // ********************************************************************************************
    if(SUCCEEDED(hr))
    {
        UINT responseCurve[10] = {0};
        // ****************************************************************************************
        // The response curve consists of an array of byte pairs.
        // The first byte contains the percentage brightness offset to be applied to the display.
        // The second byte contains the corresponding ambient light value (in LUX).
        // ****************************************************************************************
        // (0, 10)
        responseCurve[0] = 0; responseCurve[1] = 10;
        // (10, 40)
        responseCurve[2] = 10; responseCurve[3] = 40;
        // (40, 80)
        responseCurve[4] = 40; responseCurve[5] = 80;
        // (68, 100)
        responseCurve[6] = 68; responseCurve[7] = 100;
        // (90, 150)
        responseCurve[8] = 90; responseCurve[9] = 150;

        PROPVARIANT pvCurve = {0};
        InitPropVariantFromBuffer(responseCurve, 10 * sizeof (UINT), &pvCurve);
        hr = m_pSensorProperties->SetValue(SENSOR_PROPERTY_LIGHT_RESPONSE_CURVE, &pvCurve);

        PropVariantClear(&pvCurve);
    }

    if(SUCCEEDED(hr))
        hr = m_pSensorProperties->SetUnsignedIntegerValue(SENSOR_PROPERTY_CHANGE_SENSITIVITY, DEFAULT_CHANGE_SENSITIVITY);
   
    // *****************************************************************************************
    // Default values for SENSOR DATA FIELDS
    // *****************************************************************************************
    
    if(SUCCEEDED(hr))
    {
        PROPVARIANT var;
        PropVariantInit( &var );

        //Get the current time in  SYSTEMTIME format
        SYSTEMTIME st;
        ::GetSystemTime(&st);

        // Convert the SYSTEMTIME into FILETIME
        FILETIME ft;
        if(FALSE == ::SystemTimeToFileTime(&st, &ft))
        {
            hr = HRESULT_FROM_WIN32(GetLastError());
        }

        if (S_OK == hr)
        {
            var.vt                      = VT_FILETIME;
            var.filetime.dwLowDateTime  = ft.dwLowDateTime;
            var.filetime.dwHighDateTime = ft.dwHighDateTime;
            hr = m_pSensorDataFields->SetValue(SENSOR_DATA_TYPE_TIMESTAMP, &var);
        }

        PropVariantClear( &var );
    }

    if(SUCCEEDED(hr))
        hr = m_pSensorDataFields->SetFloatValue(SENSOR_DATA_TYPE_LIGHT_LEVEL_LUX, DEFAULT_LIGHT_LEVEL_LUX);

    return hr;
}

HRESULT CBadgeSensorDevice::SetSwitchArrayLeftDefaultValues()
{
    HRESULT hr = S_OK;

    if((NULL == m_pSensorProperties) || (NULL == m_pSensorDataFields))
    {
        hr = E_POINTER;
    }

    // *****************************************************************************************
    // Default values for SENSOR PROPERTIES
    // *****************************************************************************************

    if(SUCCEEDED(hr))
        hr = m_pSensorProperties->SetGuidValue(WPD_FUNCTIONAL_OBJECT_CATEGORY, SENSOR_CATEGORY_MECHANICAL);

    if(SUCCEEDED(hr))
        hr = m_pSensorProperties->SetGuidValue(SENSOR_PROPERTY_TYPE, SENSOR_TYPE_BOOLEAN_SWITCH_ARRAY);

    if(SUCCEEDED(hr))
        hr = m_pSensorProperties->SetUnsignedIntegerValue(SENSOR_PROPERTY_STATE, SENSOR_STATE_MIN);

    if(SUCCEEDED(hr))
        hr = m_pSensorProperties->SetUnsignedIntegerValue(SENSOR_PROPERTY_MIN_REPORT_INTERVAL, DEFAULT_MIN_REPORT_INTERVAL);

    if(SUCCEEDED(hr))
        hr = m_pSensorProperties->SetUnsignedIntegerValue(SENSOR_PROPERTY_CURRENT_REPORT_INTERVAL, AWAKE_CURRENT_REPORT_INTERVAL);

    if(SUCCEEDED(hr))
        hr = m_pSensorProperties->SetStringValue(SENSOR_PROPERTY_MANUFACTURER, DEVICE_MANUFACTURER_VALUE);

    if(SUCCEEDED(hr))
        hr = m_pSensorProperties->SetStringValue(SENSOR_PROPERTY_MODEL, DEVICE_MODEL_VALUE);

    if(SUCCEEDED(hr))
        hr = m_pSensorProperties->SetStringValue(SENSOR_PROPERTY_SERIAL_NUMBER, DEVICE_SERIAL_NUMBER_VALUE);

    if(SUCCEEDED(hr))
        hr = m_pSensorProperties->SetStringValue(SENSOR_PROPERTY_FRIENDLY_NAME, SENSOR_SWITCH_ARRAY_LEFT_FRIENDLY_NAME);

    if(SUCCEEDED(hr))
        hr = m_pSensorProperties->SetStringValue(SENSOR_PROPERTY_DESCRIPTION, SENSOR_SWITCH_ARRAY_LEFT_FRIENDLY_NAME);

    if(SUCCEEDED(hr))
        hr = m_pSensorProperties->SetUnsignedIntegerValue(SENSOR_PROPERTY_CONNECTION_TYPE, SENSOR_CONNECTION_TYPE_PC_EXTERNAL);

    if(SUCCEEDED(hr))
        hr = m_pSensorProperties->SetUnsignedIntegerValue(SENSOR_PROPERTY_CHANGE_SENSITIVITY, DEFAULT_CHANGE_SENSITIVITY);
   
    // *****************************************************************************************
    // Default values for SENSOR DATA FIELDS
    // *****************************************************************************************
    
    if(SUCCEEDED(hr))
    {
        PROPVARIANT var;
        PropVariantInit( &var );

        //Get the current time in  SYSTEMTIME format
        SYSTEMTIME st;
        ::GetSystemTime(&st);

        // Convert the SYSTEMTIME into FILETIME
        FILETIME ft;
        if(FALSE == ::SystemTimeToFileTime(&st, &ft))
        {
            hr = HRESULT_FROM_WIN32(GetLastError());
        }

        if (S_OK == hr)
        {
            var.vt                      = VT_FILETIME;
            var.filetime.dwLowDateTime  = ft.dwLowDateTime;
            var.filetime.dwHighDateTime = ft.dwHighDateTime;
            hr = m_pSensorDataFields->SetValue(SENSOR_DATA_TYPE_TIMESTAMP, &var);
        }

        PropVariantClear( &var );
    }

    if(SUCCEEDED(hr))
        hr = m_pSensorDataFields->SetUnsignedIntegerValue(SENSOR_DATA_TYPE_BOOLEAN_SWITCH_ARRAY_STATE, DEFAULT_BOOLEAN_SWITCH_ARRAY_STATE);

    return hr;
}


HRESULT CBadgeSensorDevice::SetSwitchArrayRightDefaultValues()
{
    HRESULT hr = S_OK;

    if((NULL == m_pSensorProperties) || (NULL == m_pSensorDataFields))
    {
        hr = E_POINTER;
    }

    // *****************************************************************************************
    // Default values for SENSOR PROPERTIES
    // *****************************************************************************************

    if(SUCCEEDED(hr))
        hr = m_pSensorProperties->SetGuidValue(WPD_FUNCTIONAL_OBJECT_CATEGORY, SENSOR_CATEGORY_MECHANICAL);

    if(SUCCEEDED(hr))
        hr = m_pSensorProperties->SetGuidValue(SENSOR_PROPERTY_TYPE, SENSOR_TYPE_BOOLEAN_SWITCH_ARRAY);

    if(SUCCEEDED(hr))
        hr = m_pSensorProperties->SetUnsignedIntegerValue(SENSOR_PROPERTY_STATE, SENSOR_STATE_MIN);

    if(SUCCEEDED(hr))
        hr = m_pSensorProperties->SetUnsignedIntegerValue(SENSOR_PROPERTY_MIN_REPORT_INTERVAL, DEFAULT_MIN_REPORT_INTERVAL);

    if(SUCCEEDED(hr))
        hr = m_pSensorProperties->SetUnsignedIntegerValue(SENSOR_PROPERTY_CURRENT_REPORT_INTERVAL, AWAKE_CURRENT_REPORT_INTERVAL);

    if(SUCCEEDED(hr))
        hr = m_pSensorProperties->SetStringValue(SENSOR_PROPERTY_MANUFACTURER, DEVICE_MANUFACTURER_VALUE);

    if(SUCCEEDED(hr))
        hr = m_pSensorProperties->SetStringValue(SENSOR_PROPERTY_MODEL, DEVICE_MODEL_VALUE);

    if(SUCCEEDED(hr))
        hr = m_pSensorProperties->SetStringValue(SENSOR_PROPERTY_SERIAL_NUMBER, DEVICE_SERIAL_NUMBER_VALUE);

    if(SUCCEEDED(hr))
        hr = m_pSensorProperties->SetStringValue(SENSOR_PROPERTY_FRIENDLY_NAME, SENSOR_SWITCH_ARRAY_RIGHT_FRIENDLY_NAME);

    if(SUCCEEDED(hr))
        hr = m_pSensorProperties->SetStringValue(SENSOR_PROPERTY_DESCRIPTION, SENSOR_SWITCH_ARRAY_RIGHT_FRIENDLY_NAME);

    if(SUCCEEDED(hr))
        hr = m_pSensorProperties->SetUnsignedIntegerValue(SENSOR_PROPERTY_CONNECTION_TYPE, SENSOR_CONNECTION_TYPE_PC_EXTERNAL);

    if(SUCCEEDED(hr))
        hr = m_pSensorProperties->SetUnsignedIntegerValue(SENSOR_PROPERTY_CHANGE_SENSITIVITY, DEFAULT_CHANGE_SENSITIVITY);
   
    // *****************************************************************************************
    // Default values for SENSOR DATA FIELDS
    // *****************************************************************************************
    
    if(SUCCEEDED(hr))
    {
        PROPVARIANT var;
        PropVariantInit( &var );

        //Get the current time in  SYSTEMTIME format
        SYSTEMTIME st;
        ::GetSystemTime(&st);

        // Convert the SYSTEMTIME into FILETIME
        FILETIME ft;
        if(FALSE == ::SystemTimeToFileTime(&st, &ft))
        {
            hr = HRESULT_FROM_WIN32(GetLastError());
        }

        if (S_OK == hr)
        {
            var.vt                      = VT_FILETIME;
            var.filetime.dwLowDateTime  = ft.dwLowDateTime;
            var.filetime.dwHighDateTime = ft.dwHighDateTime;
            hr = m_pSensorDataFields->SetValue(SENSOR_DATA_TYPE_TIMESTAMP, &var);
        }

        PropVariantClear( &var );
    }

    if(SUCCEEDED(hr))
        hr = m_pSensorDataFields->SetUnsignedIntegerValue(SENSOR_DATA_TYPE_BOOLEAN_SWITCH_ARRAY_STATE, DEFAULT_BOOLEAN_SWITCH_ARRAY_STATE);

    return hr;
}


// Sets the timestamp when new data is updated
HRESULT CBadgeSensorDevice::SetTimeStamp()
{
    CComCritSecLock<CComAutoCriticalSection> scopeLock(m_CriticalSection);

    HRESULT hr = S_OK;

    PROPVARIANT var;
    PropVariantInit( &var );

    //Get the current time in  SYSTEMTIME format
    SYSTEMTIME st;
    ::GetSystemTime(&st);

    // Convert the SYSTEMTIME into FILETIME
    FILETIME ft;
    if(FALSE == ::SystemTimeToFileTime(&st, &ft))
    {
        hr = HRESULT_FROM_WIN32(GetLastError());
    }

    if (S_OK == hr)
    {
        var.vt                      = VT_FILETIME;
        var.filetime.dwLowDateTime  = ft.dwLowDateTime;
        var.filetime.dwHighDateTime = ft.dwHighDateTime;
        hr = m_pSensorDataFields->SetValue(SENSOR_DATA_TYPE_TIMESTAMP, &var);
    }

    PropVariantClear( &var );

    return hr;
}

// Called when new data is updated from device
VOID CBadgeSensorDevice::RaiseDataEvent()
{
    CComCritSecLock<CComAutoCriticalSection> scopeLock(m_CriticalSection);

    // Check if there are any subscribers
    if( 0 < m_dwEventSubscribers )
    {
        // Set the data event flag 
        m_fValidDataEvent = TRUE;
        
        if( NULL != m_hBadgeDriverEvent )
        {
            SetEvent(m_hBadgeDriverEvent);
        }
    }
}

// Checks if there is a valid data event
// If so returns TRUE and resets the internal flag
BOOL CBadgeSensorDevice::HasValidDataEvent()
{
    CComCritSecLock<CComAutoCriticalSection> scopeLock(m_CriticalSection);

    BOOL fDataEvent = FALSE;

    // Check if there is a valid data event to post
    if(TRUE == m_fValidDataEvent)
    {
        fDataEvent = m_fValidDataEvent;
        m_fValidDataEvent = FALSE;
    }

    return fDataEvent;
}

// Sets the persistent unique ID property
HRESULT CBadgeSensorDevice::SetUniqueID(__in IWDFDevice* pWdfDevice)
{
    HRESULT hr = S_OK;

    CComPtr<IWDFNamedPropertyStore> spPropStore;
    if (SUCCEEDED(hr))
    {
        hr = pWdfDevice->RetrieveDevicePropertyStore(NULL, WdfPropertyStoreCreateIfMissing, &spPropStore, NULL);
    }

    if(SUCCEEDED(hr))
    {
        GUID idGuid;
        
        LPCWSTR lpcszKeyName = GetSensorObjectID();
        
        PROPVARIANT vID;
        hr = spPropStore->GetNamedValue(lpcszKeyName, &vID);
        if (SUCCEEDED(hr))
        {
            hr = ::CLSIDFromString(vID.bstrVal, &idGuid);
        }
        else
        {
            hr = ::CoCreateGuid(&idGuid);
            if (SUCCEEDED(hr))
            {
                LPOLESTR lpszGUID = NULL;
                hr = ::StringFromCLSID(idGuid, &lpszGUID);
                if (SUCCEEDED(hr))
                {
                    PropVariantInit(&vID);
                    vID.vt = VT_LPWSTR;
                    vID.pwszVal = lpszGUID;
                    hr = spPropStore->SetNamedValue(lpcszKeyName, &vID);
                    PropVariantClear(&vID);
                }
            }
        }

        if (SUCCEEDED(hr))
        {
            hr = m_pSensorProperties->SetGuidValue(SENSOR_PROPERTY_PERSISTENT_UNIQUE_ID, idGuid);
        }
    }

    return hr;
}